@@ -3,46 +3,50 @@ package IkiWiki::Plugin::conditional;
 
 use warnings;
 use strict;
-use IkiWiki 2.00;
+use IkiWiki 3.00;
 use UNIVERSAL;
 
-sub import { #{{{
+sub import {
+	hook(type => "getsetup", id => "conditional", call => \&getsetup);
 	hook(type => "preprocess", id => "if", call => \&preprocess_if);
-} # }}}
+}
 
-sub preprocess_if (@) { #{{{
+sub getsetup {
+	return
+		plugin => {
+			safe => 1,
+			rebuild => undef,
+			section => "widget",
+		},
+}
+
+sub preprocess_if (@) {
 	my %params=@_;
 
 	foreach my $param (qw{test then}) {
 		if (! exists $params{$param}) {
-			return "[[if ".sprintf(gettext('%s parameter is required'), $param)."]]";
+			error sprintf(gettext('%s parameter is required'), $param);
 		}
 	}
 
 	my $result=0;
-	if ((exists $params{all} && lc $params{all} eq "no") ||
-		# An optimisation to avoid needless looping over every page
-		# and adding of dependencies for simple uses of some of the
-		# tests.
-		$params{test} =~ /^(enabled|sourcepage|destpage)\((.*)\)$/) {
-		add_depends($params{page}, "$params{test} and $params{page}");
+	if ((exists $params{all} && ! IkiWiki::yesno($params{all})) ||
+	    # An optimisation to avoid needless looping over every page
+	    # for simple uses of some of the tests.
+	    $params{test} =~ /^([\s\!()]*((enabled|sourcepage|destpage|included)\([^)]*\)|(and|or))[\s\!()]*)+$/) {
+		add_depends($params{page}, "($params{test}) and $params{page}");
 		$result=pagespec_match($params{page}, $params{test},
 				location => $params{page},
 				sourcepage => $params{page},
 				destpage => $params{destpage});
 	}
 	else {
-		add_depends($params{page}, $params{test});
-
-		foreach my $page (keys %pagesources) {
-			if (pagespec_match($page, $params{test}, 
-					location => $params{page},
-					sourcepage => $params{page},
-					destpage => $params{destpage})) {
-				$result=1;
-				last;
-			}
-		}
+		$result=pagespec_match_list($params{page}, $params{test},
+			# stop after first match
+			num => 1,
+			sourcepage => $params{page},
+			destpage => $params{destpage},
+		);
 	}
 
 	my $ret;
@@ -55,13 +59,12 @@ sub preprocess_if (@) { #{{{
 	else {
 		$ret="";
 	}
-	return IkiWiki::preprocess($params{page}, $params{destpage}, 
-		IkiWiki::filter($params{page}, $params{destpage}, $ret));
-} # }}}
+	return IkiWiki::preprocess($params{page}, $params{destpage}, $ret);
+}
 
 package IkiWiki::PageSpec;
 
-sub match_enabled ($$;@) { #{{{
+sub match_enabled ($$;@) {
 	shift;
 	my $plugin=shift;
 	
@@ -72,12 +75,14 @@ sub match_enabled ($$;@) { #{{{
 	else {
 		return IkiWiki::FailReason->new("$plugin is not enabled");
 	}
-} #}}}
+}
 
-sub match_sourcepage ($$;@) { #{{{
+sub match_sourcepage ($$;@) {
 	shift;
 	my $glob=shift;
 	my %params=@_;
+	
+	$glob=derel($glob, $params{location});
 
 	return IkiWiki::FailReason->new("cannot match sourcepage") unless exists $params{sourcepage};
 	if (match_glob($params{sourcepage}, $glob, @_)) {
@@ -86,13 +91,15 @@ sub match_sourcepage ($$;@) { #{{{
 	else {
 		return IkiWiki::FailReason->new("sourcepage does not match $glob");
 	}
-} #}}}
+}
 
-sub match_destpage ($$;@) { #{{{
+sub match_destpage ($$;@) {
 	shift;
 	my $glob=shift;
 	my %params=@_;
 	
+	$glob=derel($glob, $params{location});
+
 	return IkiWiki::FailReason->new("cannot match destpage") unless exists $params{destpage};
 	if (match_glob($params{destpage}, $glob, @_)) {
 		return IkiWiki::SuccessReason->new("destpage matches $glob");
@@ -100,9 +107,9 @@ sub match_destpage ($$;@) { #{{{
 	else {
 		return IkiWiki::FailReason->new("destpage does not match $glob");
 	}
-} #}}}
+}
 
-sub match_included ($$;$) { #{{{
+sub match_included ($$;@) {
 	shift;
 	shift;
 	my %params=@_;
@@ -114,6 +121,6 @@ sub match_included ($$;$) { #{{{
 	else {
 		return IkiWiki::FailReason->new("page $params{sourcepage} is not included");
 	}
-} #}}}
+}
 
 1
