@@ -4,17 +4,27 @@ package IkiWiki::Plugin::meta;
 
 use warnings;
 use strict;
-use IkiWiki 2.00;
+use IkiWiki 3.00;
 
 my %metaheaders;
 
-sub import { #{{{
+sub import {
+	hook(type => "getsetup", id => "meta", call => \&getsetup);
 	hook(type => "needsbuild", id => "meta", call => \&needsbuild);
 	hook(type => "preprocess", id => "meta", call => \&preprocess, scan => 1);
 	hook(type => "pagetemplate", id => "meta", call => \&pagetemplate);
-} # }}}
+}
 
-sub needsbuild (@) { #{{{
+sub getsetup () {
+	return
+		plugin => {
+			safe => 1,
+			rebuild => undef,
+			section => "core",
+		},
+}
+
+sub needsbuild (@) {
 	my $needsbuild=shift;
 	foreach my $page (keys %pagestate) {
 		if (exists $pagestate{$page}{meta}) {
@@ -29,16 +39,17 @@ sub needsbuild (@) { #{{{
 	}
 }
 
-sub scrub ($) { #{{{
+sub scrub ($$$) {
 	if (IkiWiki::Plugin::htmlscrubber->can("sanitize")) {
-		return IkiWiki::Plugin::htmlscrubber::sanitize(content => shift);
+		return IkiWiki::Plugin::htmlscrubber::sanitize(
+			content => shift, page => shift, destpage => shift);
 	}
 	else {
 		return shift;
 	}
-} #}}}
+}
 
-sub safeurl ($) { #{{{
+sub safeurl ($) {
 	my $url=shift;
 	if (exists $IkiWiki::Plugin::htmlscrubber::{safe_url_regexp} &&
 	    defined $IkiWiki::Plugin::htmlscrubber::safe_url_regexp) {
@@ -47,9 +58,9 @@ sub safeurl ($) { #{{{
 	else {
 		return 1;
 	}
-} #}}}
+}
 
-sub htmlize ($$$) { #{{{
+sub htmlize ($$$) {
 	my $page = shift;
 	my $destpage = shift;
 
@@ -58,7 +69,7 @@ sub htmlize ($$$) { #{{{
 		IkiWiki::preprocess($page, $destpage, shift)));
 }
 
-sub preprocess (@) { #{{{
+sub preprocess (@) {
 	return "" unless @_;
 	my %params=@_;
 	my $key=shift;
@@ -77,11 +88,21 @@ sub preprocess (@) { #{{{
 
 	# Metadata collection that needs to happen during the scan pass.
 	if ($key eq 'title') {
-		$pagestate{$page}{meta}{title}=HTML::Entities::encode_numeric($value);
-		# fallthrough
+		$pagestate{$page}{meta}{title}=$value;
+		if (exists $params{sortas}) {
+			$pagestate{$page}{meta}{titlesort}=$params{sortas};
+		}
+		else {
+			delete $pagestate{$page}{meta}{titlesort};
+		}
+		return "";
 	}
 	elsif ($key eq 'description') {
-		$pagestate{$page}{meta}{description}=HTML::Entities::encode_numeric($value);
+		$pagestate{$page}{meta}{description}=$value;
+		# fallthrough
+	}
+	elsif ($key eq 'guid') {
+		$pagestate{$page}{meta}{guid}=$value;
 		# fallthrough
 	}
 	elsif ($key eq 'license') {
@@ -96,35 +117,51 @@ sub preprocess (@) { #{{{
 	}
 	elsif ($key eq 'link' && ! %params) {
 		# hidden WikiLink
-		push @{$links{$page}}, $value;
+		add_link($page, $value);
 		return "";
 	}
 	elsif ($key eq 'author') {
 		$pagestate{$page}{meta}{author}=$value;
+		if (exists $params{sortas}) {
+			$pagestate{$page}{meta}{authorsort}=$params{sortas};
+		}
+		else {
+			delete $pagestate{$page}{meta}{authorsort};
+		}
 		# fallthorough
 	}
 	elsif ($key eq 'authorurl') {
 		$pagestate{$page}{meta}{authorurl}=$value if safeurl($value);
 		# fallthrough
 	}
+	elsif ($key eq 'permalink') {
+		$pagestate{$page}{meta}{permalink}=$value if safeurl($value);
+		# fallthrough
+	}
+	elsif ($key eq 'date') {
+		eval q{use Date::Parse};
+		if (! $@) {
+			my $time = str2time($value);
+			$IkiWiki::pagectime{$page}=$time if defined $time;
+		}
+	}
+	elsif ($key eq 'updated') {
+		eval q{use Date::Parse};
+		if (! $@) {
+			my $time = str2time($value);
+			$pagestate{$page}{meta}{updated}=$time if defined $time;
+		}
+	}
 
 	if (! defined wantarray) {
 		# avoid collecting duplicate data during scan pass
 		return;
 	}
 
-	# Metadata collection that happens only during preprocessing pass.
-	if ($key eq 'date') {
-		eval q{use Date::Parse};
-		if (! $@) {
-			my $time = str2time($value);
-			$IkiWiki::pagectime{$page}=$time if defined $time;
-		}
-	}
-	elsif ($key eq 'permalink') {
+	# Metadata handling that happens only during preprocessing pass.
+	if ($key eq 'permalink') {
 		if (safeurl($value)) {
-			$pagestate{$page}{meta}{permalink}=$value;
-			push @{$metaheaders{$page}}, scrub('<link rel="bookmark" href="'.encode_entities($value).'" />');
+			push @{$metaheaders{$page}}, scrub('<link rel="bookmark" href="'.encode_entities($value).'" />', $page, $destpage);
 		}
 	}
 	elsif ($key eq 'stylesheet') {
@@ -134,25 +171,30 @@ sub preprocess (@) { #{{{
 		# editable page as a stylesheet
 		my $stylesheet=bestlink($page, $value.".css");
 		if (! length $stylesheet) {
-			return "[[meta ".gettext("stylesheet not found")."]]";
+			error gettext("stylesheet not found")
 		}
-		push @{$metaheaders{$page}}, '<link href="'.urlto($stylesheet, $page).
+		push @{$metaheaders{$page}}, scrub('<link href="'.urlto($stylesheet, $page).
 			'" rel="'.encode_entities($rel).
 			'" title="'.encode_entities($title).
-			"\" type=\"text/css\" />";
+			"\" type=\"text/css\" />", $page, $destpage);
 	}
 	elsif ($key eq 'openid') {
+		my $delegate=0; # both by default
+		if (exists $params{delegate}) {
+			$delegate = 1 if lc $params{delegate} eq 'openid';
+			$delegate = 2 if lc $params{delegate} eq 'openid2';
+		}
 		if (exists $params{server} && safeurl($params{server})) {
 			push @{$metaheaders{$page}}, '<link href="'.encode_entities($params{server}).
-				'" rel="openid.server" />';
+				'" rel="openid.server" />' if $delegate ne 2;
 			push @{$metaheaders{$page}}, '<link href="'.encode_entities($params{server}).
-				'" rel="openid2.provider" />';
+				'" rel="openid2.provider" />' if $delegate ne 1;
 		}
 		if (safeurl($value)) {
 			push @{$metaheaders{$page}}, '<link href="'.encode_entities($value).
-				'" rel="openid.delegate" />';
+				'" rel="openid.delegate" />' if $delegate ne 2;
 			push @{$metaheaders{$page}}, '<link href="'.encode_entities($value).
-				'" rel="openid2.local_id" />';
+				'" rel="openid2.local_id" />' if $delegate ne 1;
 		}
 		if (exists $params{"xrds-location"} && safeurl($params{"xrds-location"})) {
 			push @{$metaheaders{$page}}, '<meta http-equiv="X-XRDS-Location"'.
@@ -165,11 +207,11 @@ sub preprocess (@) { #{{{
 		if ($value !~ /^\w+:\/\//) {
 			my ($redir_page, $redir_anchor) = split /\#/, $value;
 
-			add_depends($page, $redir_page);
 			my $link=bestlink($page, $redir_page);
 			if (! length $link) {
-				return "[[meta ".gettext("redir page not found")."]]";
+				error gettext("redir page not found")
 			}
+			add_depends($page, $link, deptype("presence"));
 
 			$value=urlto($link, $page);
 			$value.='#'.$redir_anchor if defined $redir_anchor;
@@ -181,7 +223,7 @@ sub preprocess (@) { #{{{
 			my %seen;
 			while (exists $pagestate{$at}{meta}{redir}) {
 				if ($seen{$at}) {
-					return "[[meta ".gettext("redir cycle is not allowed")."]]";
+					error gettext("redir cycle is not allowed")
 				}
 				$seen{$at}=1;
 				$at=$pagestate{$at}{meta}{redir};
@@ -193,7 +235,7 @@ sub preprocess (@) { #{{{
 		my $delay=int(exists $params{delay} ? $params{delay} : 0);
 		my $redir="<meta http-equiv=\"refresh\" content=\"$delay; URL=$value\" />";
 		if (! $safe) {
-			$redir=scrub($redir);
+			$redir=scrub($redir, $page, $destpage);
 		}
 		push @{$metaheaders{$page}}, $redir;
 	}
@@ -203,22 +245,34 @@ sub preprocess (@) { #{{{
 				join(" ", map {
 					encode_entities($_)."=\"".encode_entities(decode_entities($params{$_}))."\""
 				} keys %params).
-				" />\n");
+				" />\n", $page, $destpage);
 		}
 	}
 	elsif ($key eq 'robots') {
 		push @{$metaheaders{$page}}, '<meta name="robots"'.
 			' content="'.encode_entities($value).'" />';
 	}
+	elsif ($key eq 'description') {
+		push @{$metaheaders{$page}}, '<meta name="'.
+			encode_entities($key).
+			'" content="'.encode_entities($value).'" />';
+	}
+	elsif ($key eq 'name') {
+		push @{$metaheaders{$page}}, scrub('<meta '.$key.'="'.
+			encode_entities($value).
+			join(' ', map { "$_=\"$params{$_}\"" } keys %params).
+			' />', $page, $destpage);
+	}
 	else {
-		push @{$metaheaders{$page}}, scrub('<meta name="'.encode_entities($key).
-			'" content="'.encode_entities($value).'" />');
+		push @{$metaheaders{$page}}, scrub('<meta name="'.
+			encode_entities($key).'" content="'.
+			encode_entities($value).'" />', $page, $destpage);
 	}
 
 	return "";
-} # }}}
+}
 
-sub pagetemplate (@) { #{{{
+sub pagetemplate (@) {
 	my %params=@_;
         my $page=$params{page};
         my $destpage=$params{destpage};
@@ -230,7 +284,7 @@ sub pagetemplate (@) { #{{{
 		$template->param(meta => join("\n", grep { (! $seen{$_}) && ($seen{$_}=1) } @{$metaheaders{$page}}));
 	}
 	if (exists $pagestate{$page}{meta}{title} && $template->query(name => "title")) {
-		$template->param(title => $pagestate{$page}{meta}{title});
+		$template->param(title => HTML::Entities::encode_numeric($pagestate{$page}{meta}{title}));
 		$template->param(title_overridden => 1);
 	}
 
@@ -239,6 +293,11 @@ sub pagetemplate (@) { #{{{
 			if exists $pagestate{$page}{meta}{$field} && $template->query(name => $field);
 	}
 
+	foreach my $field (qw{description}) {
+		$template->param($field => HTML::Entities::encode_numeric($pagestate{$page}{meta}{$field}))
+			if exists $pagestate{$page}{meta}{$field} && $template->query(name => $field);
+	}
+
 	foreach my $field (qw{license copyright}) {
 		if (exists $pagestate{$page}{meta}{$field} && $template->query(name => $field) &&
 		    ($page eq $destpage || ! exists $pagestate{$destpage}{meta}{$field} ||
@@ -246,9 +305,36 @@ sub pagetemplate (@) { #{{{
 			$template->param($field => htmlize($page, $destpage, $pagestate{$page}{meta}{$field}));
 		}
 	}
-} # }}}
+}
 
-sub match { #{{{
+sub get_sort_key {
+	my $page = shift;
+	my $meta = shift;
+
+	# e.g. titlesort (also makes sense for author)
+	my $key = $pagestate{$page}{meta}{$meta . "sort"};
+	return $key if defined $key;
+
+	# e.g. title
+	$key = $pagestate{$page}{meta}{$meta};
+	return $key if defined $key;
+
+	# fall back to closer-to-core things
+	if ($meta eq 'title') {
+		return pagetitle(IkiWiki::basename($page));
+	}
+	elsif ($meta eq 'date') {
+		return $IkiWiki::pagectime{$page};
+	}
+	elsif ($meta eq 'updated') {
+		return $IkiWiki::pagemtime{$page};
+	}
+	else {
+		return '';
+	}
+}
+
+sub match {
 	my $field=shift;
 	my $page=shift;
 	
@@ -260,42 +346,69 @@ sub match { #{{{
 		$val=$pagestate{$page}{meta}{$field};
 	}
 	elsif ($field eq 'title') {
-		$val=pagetitle($page);
+		$val = pagetitle($page);
 	}
 
 	if (defined $val) {
 		if ($val=~/^$re$/i) {
-			return IkiWiki::SuccessReason->new("$re matches $field of $page");
+			return IkiWiki::SuccessReason->new("$re matches $field of $page", $page => $IkiWiki::DEPEND_CONTENT, "" => 1);
 		}
 		else {
-			return IkiWiki::FailReason->new("$re does not match $field of $page");
+			return IkiWiki::FailReason->new("$re does not match $field of $page", $page => $IkiWiki::DEPEND_CONTENT, "" => 1);
 		}
 	}
 	else {
-		return IkiWiki::FailReason->new("$page does not have a $field");
+		return IkiWiki::FailReason->new("$page does not have a $field", $page => $IkiWiki::DEPEND_CONTENT);
 	}
-} #}}}
+}
 
 package IkiWiki::PageSpec;
 
-sub match_title ($$;@) { #{{{
-	IkiWiki::Plugin::meta::match("title", @_);	
-} #}}}
+sub match_title ($$;@) {
+	IkiWiki::Plugin::meta::match("title", @_);
+}
 
-sub match_author ($$;@) { #{{{
+sub match_author ($$;@) {
 	IkiWiki::Plugin::meta::match("author", @_);
-} #}}}
+}
 
-sub match_authorurl ($$;@) { #{{{
+sub match_authorurl ($$;@) {
 	IkiWiki::Plugin::meta::match("authorurl", @_);
-} #}}}
+}
 
-sub match_license ($$;@) { #{{{
+sub match_license ($$;@) {
 	IkiWiki::Plugin::meta::match("license", @_);
-} #}}}
+}
 
-sub match_copyright ($$;@) { #{{{
+sub match_copyright ($$;@) {
 	IkiWiki::Plugin::meta::match("copyright", @_);
-} #}}}
+}
+
+sub match_guid ($$;@) {
+	IkiWiki::Plugin::meta::match("guid", @_);
+}
+
+package IkiWiki::SortSpec;
+
+sub cmp_meta {
+	my $meta = shift;
+	error(gettext("sort=meta requires a parameter")) unless defined $meta;
+
+	if ($meta eq 'updated' || $meta eq 'date') {
+		return IkiWiki::Plugin::meta::get_sort_key($a, $meta)
+			<=>
+			IkiWiki::Plugin::meta::get_sort_key($b, $meta);
+	}
+
+	return IkiWiki::Plugin::meta::get_sort_key($a, $meta)
+		cmp
+		IkiWiki::Plugin::meta::get_sort_key($b, $meta);
+}
+
+# A prototype of how sort=title could behave in 4.0 or something
+sub cmp_meta_title {
+	$_[0] = 'title';
+	return cmp_meta(@_);
+}
 
 1
